import {
  DEFAULT_TARGET_RADIUS, DISTANCE_FACTOR, MINIMUM_FORCE, OVERLAP_FORCE, VISCOSITY_FACTOR,
} from '../../util/constants.js';
import { Box } from './box.js';
import { Rectangle, Vector } from './geometry.js';

// Render children with absolute css positioning.

// Let there be a force between elements such that the force between
// any two elements is along the line between their centers,
// so that the elements repel when too close but attract when too far.

// The equilibrium distance can be tuned, e.g. can be scaled by an input.

// NOTE: (with optional overlay preferring a grid or some other shape?),

// NOTE: Could also allow user input dragging elements.
// What might be neat here is to implement a force-based resistance effect;
// basically, the mouse pointer drags the element with a spring rather than directly.
// If the shape of the graph resists the transformation,
// the distance between the element and the cursor should increase.

// On an interval, compute forces among the elements.
// Simulate the effects of these forces

// NOTE: Impart random nudges, and resolve their effects to a user-visible resolution
// before rendering.

// NOTE: When mouse is in our box, we could hijack the scroll actions to zoom in/out.

export class ForceDirectedGraph extends Box {
  constructor(name, parentEl, options = {}) {
    super(name, parentEl, options);
    this.addClass('fixed');
  }

  addBox(name) {
    const box = super.addBox(name);
    box.addClass('absolute');
    box.el.style.left = '0px';
    box.el.style.top = '0px';
    box.velocity = Vector.from([0, 0]);
    return box;
  }

  static pairwiseForce(boxA, boxB, targetRadius) {
    const rectA = boxA instanceof Rectangle ? boxA : boxA.getGeometry();
    const centerA = rectA.center;
    const rectB = boxB instanceof Rectangle ? boxB : boxB.getGeometry();
    const centerB = rectB.center;
    const r = centerB.subtract(centerA);

    // Apply a stronger force when overlap occurs
    if (rectA.doesOverlap(rectB)) {
      // if their centers actually coincide we can just randomize the direction.
      if (r.magnitudeSquared === 0) {
        return Vector.randomUnitVector(rectA.dim).scale(OVERLAP_FORCE);
      }
      return r.normalize().scale(OVERLAP_FORCE);
    }
    // repel if closer than targetRadius
    // attract if farther than targetRadius
    const force = -DISTANCE_FACTOR * (r.magnitude - targetRadius);
    return r.normalize().scale(force);
  }

  computeEulerFrame(tDelta) {
    // Compute all net forces
    const netForces = Array.from(Array(this.boxes.length), () => Vector.from([0, 0]));
    for (const boxA of this.boxes) {
      const idxA = this.boxes.indexOf(boxA);
      for (const boxB of this.boxes.slice(idxA + 1)) {
        const idxB = this.boxes.indexOf(boxB);
        const force = ForceDirectedGraph.pairwiseForce(boxA, boxB, DEFAULT_TARGET_RADIUS);
        // Ignore forces below a certain threshold
        if (force.magnitude >= MINIMUM_FORCE) {
          netForces[idxA] = netForces[idxA].subtract(force);
          netForces[idxB] = netForces[idxB].add(force);
        }
      }
    }

    // Compute motions
    for (const box of this.boxes) {
      const idx = this.boxes.indexOf(box);
      box.velocity = box.velocity.add(netForces[idx].scale(tDelta));
      // Apply some drag
      box.velocity = box.velocity.scale(1 - VISCOSITY_FACTOR);
    }

    for (const box of this.boxes) {
      box.move(box.velocity);
    }

    // TODO: translate everything to keep coordinates positive
    const translate = Vector.zeros(2);
    for (const box of this.boxes) {
      const rect = box.getGeometry();
      console.log({ box, rect });
      for (const vertex of rect.vertices) {
        for (let dim = 0; dim < vertex.dim; dim++) {
          translate[dim] = Math.max(translate[dim], -vertex[dim]);
          console.log(`vertex[${dim}] = ${vertex[dim]}, translate[${dim}] = ${translate[dim]}`);
        }
      }
    }
    for (const box of this.boxes) {
      box.move(translate);
    }
  }
}
