export class Vector extends Array {
  get dim() {
    return this.length ?? 0;
  }

  add(vector) {
    if (vector.dim !== this.dim) {
      throw new Error('Can only add vectors of the same dimensions');
    }
    return Vector.from(this.map((q, idx) => q + vector[idx]));
  }

  subtract(vector) {
    if (vector.dim !== this.dim) {
      throw new Error('Can only subtract vectors of the same dimensions');
    }
    return Vector.from(this.map((q, idx) => q - vector[idx]));
  }

  static unitVector(dim, totalDim) {
    return Vector.from(Array(totalDim).map((_, idx) => (idx === dim ? 1 : 0)));
  }

  get magnitudeSquared() {
    return this.reduce((total, q) => total += q ** 2, 0);
  }

  get magnitude() {
    return Math.sqrt(this.magnitudeSquared);
  }

  scale(factor) {
    return Vector.from(this.map((q) => q * factor));
  }

  normalize() {
    return this.scale(1 / this.magnitude);
  }

  static randomUnitVector(totalDim) {
    return Vector.from(Array(totalDim), () => Math.random()).normalize();
  }

  static zeros(totalDim) {
    return Vector.from(Array(totalDim), () => 0);
  }
}

export class Polygon {
  constructor() {
    this.vertices = [];
    this.dim = 0;
  }

  addVertex(point) {
    point = point instanceof Vector ? point : Vector.from(point);
    if (!this.dim) {
      this.dim = point.dim;
    } else if (this.dim !== point.dim) {
      throw new Error('All vertices of a polygon must have the same dimensionality');
    }
    this.vertices.push(point);
  }
}

export class Rectangle extends Polygon {
  constructor(startPoint, dimensions) {
    super();
    this.startPoint = Vector.from(startPoint);
    this.dimensions = Vector.from(dimensions);
    // Next point is obtained by moving the specified length along each dimension
    // one at a time, then reversing these movements in the same order.
    let point = this.startPoint;
    for (let dim = 0; dim < dimensions.length; dim++) {
      this.addVertex(point);
      const increment = Vector.unitVector(dim, dimensions.length);
      point = point.add(increment);
    }
    for (let dim = 0; dim < dimensions.length; dim++) {
      this.addVertex(point);
      const increment = Vector.unitVector(dim, dimensions.length);
      point = point.subtract(increment);
    }
  }

  get center() {
    return Vector.from(this.dimensions.map((Q, idx) => this.startPoint[idx] + Q / 2));
  }

  doesOverlap(rect) {
    return this.dimensions.every((_, idx) => {
      const thisMin = this.startPoint[idx];
      const thisMax = this.startPoint[idx] + this.dimensions[idx];
      const thatMin = rect.startPoint[idx];
      const thatMax = rect.startPoint[idx] + rect.dimensions[idx];
      return (thisMin <= thatMin && thisMax >= thatMin)
        || (thisMin >= thatMin && thisMin <= thatMax);
    });
  }
}
